package me.unurled.sacredrealms.sr.commands.admin;

import static me.unurled.sacredrealms.sr.utils.Logger.log;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map.Entry;
import me.unurled.sacredrealms.sr.components.clientbuild.ClientBuild;
import me.unurled.sacredrealms.sr.components.clientbuild.ClientBuildManager;
import me.unurled.sacredrealms.sr.managers.Manager;
import org.bukkit.Location;
import org.bukkit.block.Block;
import org.bukkit.block.data.BlockData;
import org.bukkit.command.Command;
import org.bukkit.command.CommandSender;
import org.bukkit.command.TabExecutor;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.Listener;
import org.bukkit.event.block.BlockBreakEvent;
import org.bukkit.event.block.BlockPlaceEvent;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

public class ClientBuildCommand implements TabExecutor, Listener {

  public static final String USAGE =
      "Usage: /clientbuild <create|delete|modify|save|display> <name>";
  public static final String DOES_NOT_EXIST = "A client build with that name does not exist.";
  public static final String MODIFY = "modify";
  public static final String DISPLAY = "display";
  public static final String DELETE = "delete";
  private List<String> nameCompletions = new ArrayList<>();
  private ClientBuildManager manager = null;

  private static void display(
      @NotNull CommandSender sender,
      @NotNull String @NotNull [] args,
      Player p,
      @NotNull ClientBuildManager clientBuildManager) {
    String name = args[1].toLowerCase();
    if (!clientBuildManager.getBuildNames().contains(name)) {
      sender.sendMessage(DOES_NOT_EXIST);
      return;
    }
    ClientBuild build = clientBuildManager.getBuild(name);
    if (build == null) {
      sender.sendMessage("An error occurred while displaying the client build.");
      return;
    }
    if (clientBuildManager.getPlayerBuilds(p).contains(name)) {
      clientBuildManager.removeDisplayFromPlayer(p, build);
      sender.sendMessage("Removed the client build display.");
    } else {
      clientBuildManager.displayToPlayer(p, build);
      sender.sendMessage("Displayed the client build.");
    }
  }

  private static void save(
      @NotNull CommandSender sender,
      @NotNull String @NotNull [] args,
      Player p,
      @NotNull ClientBuildManager clientBuildManager) {
    // save the client build to the list of builds so stopping the listener.
    ClientBuild build = new ClientBuild(args[1].toLowerCase());
    if (!clientBuildManager.hasTemporaryBlocks(p)) {
      sender.sendMessage("You do not have any temporary blocks to save.");
      return;
    }
    HashMap<Location, BlockData> blocks =
        (HashMap<Location, BlockData>) clientBuildManager.getTemporaryBlocks(p);
    log("blocks: " + blocks.size());
    for (Entry<Location, BlockData> block : blocks.entrySet()) {
      build.addBlock(block.getKey(), block.getValue());
    }

    clientBuildManager.addBuild(build);

    // remove blocks from the world
    for (Entry<Location, BlockData> block :
        clientBuildManager.getTemporaryPreviousBlocks(p).entrySet()) {
      block.getKey().getWorld().setBlockData(block.getKey(), block.getValue());
    }

    clientBuildManager.removeTemporaryBlocks(p);
    clientBuildManager.removeTemporaryPreviousBlocks(p);

    sender.sendMessage("Client builds saved.");
  }

  private static void modify(
      @NotNull CommandSender sender,
      @NotNull String @NotNull [] args,
      Player p,
      @NotNull ClientBuildManager clientBuildManager) {
    String name = args[1].toLowerCase();
    if (!clientBuildManager.getBuildNames().contains(name)) {
      sender.sendMessage(DOES_NOT_EXIST);
      return;
    }
    sender.sendMessage("Modify the client build and save after you are finished.");
    clientBuildManager.setTemporaryBlocks(p, new HashMap<>());
    clientBuildManager.setTemporaryPreviousBlocks(p, new HashMap<>());
  }

  private static void delete(
      @NotNull CommandSender sender,
      @NotNull String @NotNull [] args,
      @NotNull ClientBuildManager clientBuildManager) {
    String name = args[1].toLowerCase();
    if (!clientBuildManager.getBuildNames().contains(name)) {
      sender.sendMessage(DOES_NOT_EXIST);
      return;
    }
    clientBuildManager.removeBuild(name);
    sender.sendMessage("Client build deleted.");
  }

  private static void create(
      @NotNull CommandSender sender,
      @NotNull String @NotNull [] args,
      Player p,
      @NotNull ClientBuildManager clientBuildManager) {
    String name = args[1].toLowerCase();
    if (clientBuildManager.getBuildNames().contains(name)) {
      sender.sendMessage("A client build with that name already exists.");
      return;
    }
    sender.sendMessage(
        "Client build created, place blocks that you want to add to the Client"
            + " Build and save after you are finished.");
    clientBuildManager.setTemporaryBlocks(p, new HashMap<>());
    clientBuildManager.setTemporaryPreviousBlocks(p, new HashMap<>());
  }

  private static void list(
      @NotNull CommandSender sender, @NotNull ClientBuildManager clientBuildManager) {
    List<ClientBuild> builds = clientBuildManager.getBuilds();
    if (builds.isEmpty()) {
      sender.sendMessage("There are no client builds.");
      return;
    }

    sender.sendMessage("Client Builds:");
    for (ClientBuild build : builds) {
      if (build != null) sender.sendMessage(build.getName());
    }
  }

  /**
   * Executes the given command, returning its success. <br>
   * If false is returned, then the "usage" plugin.yml entry for this command (if defined) will be
   * sent to the player.
   *
   * @param sender Source of the command
   * @param command Command which was executed
   * @param label Alias of the command which was used
   * @param args Passed command arguments
   * @return true if a valid command, otherwise false
   */
  @Override
  public boolean onCommand(
      @NotNull CommandSender sender,
      @NotNull Command command,
      @NotNull String label,
      @NotNull String[] args) {
    if (!sender.hasPermission("sr.clientbuild")) {
      sender.sendMessage("You do not have permission to use this command.");
      return true;
    }

    if (!(sender instanceof Player p)) {
      sender.sendMessage("You must be a player to use this command.");
      return true;
    }

    if (args.length == 0) {
      sender.sendMessage("Usage: /clientbuild <create|delete|list|modify|save|display>");
      return true;
    }

    ClientBuildManager clientBuildManager = Manager.getInstance(ClientBuildManager.class);
    if (args.length == 1) {
      if (args[0].equalsIgnoreCase("list")) {
        list(sender, clientBuildManager);
        return true;
      }
      sender.sendMessage(USAGE);
      return true;
    } else if (args.length == 2) {
      // create, delete, modify, save, display
      if (args[0].equalsIgnoreCase("create")) {
        create(sender, args, p, clientBuildManager);
        return true;
      } else if (args[0].equalsIgnoreCase(DELETE)) {
        delete(sender, args, clientBuildManager);
        return true;
      } else if (args[0].equalsIgnoreCase(MODIFY)) {
        modify(sender, args, p, clientBuildManager);
        return true;
      } else if (args[0].equalsIgnoreCase("save")) {
        save(sender, args, p, clientBuildManager);
        return true;
      } else if (args[0].equalsIgnoreCase(DISPLAY)) {
        display(sender, args, p, clientBuildManager);
        return true;
      }
      sender.sendMessage(USAGE);
      return true;
    }
    sender.sendMessage(USAGE);
    return true;
  }

  @EventHandler
  public void onBlockPlace(@NotNull BlockPlaceEvent e) {
    Player p = e.getPlayer();
    if (manager == null) {
      manager = Manager.getInstance(ClientBuildManager.class);
    }
    if (manager.hasTemporaryBlocks(p)) {
      log("block placed " + e.getBlock().getType());
      manager.getTemporaryBlocks(p).put(e.getBlock().getLocation(), e.getBlock().getBlockData());
    }
    if (manager.hasTemporaryPreviousBlocks(p)) {
      Block block = e.getBlock();
      // set blockstate to block
      manager
          .getTemporaryPreviousBlocks(p)
          .put(block.getLocation(), e.getBlockReplacedState().getBlockData());
    }
  }

  @EventHandler
  public void onBlockBreak(@NotNull BlockBreakEvent e) {
    Player p = e.getPlayer();
    if (manager == null) {
      manager = Manager.getInstance(ClientBuildManager.class);
    }
    if (manager.hasTemporaryBlocks(p)) {
      manager.getTemporaryBlocks(p).remove(e.getBlock().getLocation());
    }
    if (manager.hasTemporaryPreviousBlocks(p)) {
      Block block = e.getBlock();
      // set blockstate to block
      manager.getTemporaryPreviousBlocks(p).put(block.getLocation(), block.getBlockData());
    }
  }

  /**
   * Requests a list of possible completions for a command argument.
   *
   * @param sender Source of the command. For players tab-completing a command inside a command
   *     block, this will be the player, not the command block.
   * @param command Command which was executed
   * @param label Alias of the command which was used
   * @param args The arguments passed to the command, including final partial argument to be
   *     completed
   * @return A List of possible completions for the final argument, or null to default to the
   *     command executor
   */
  @Override
  public @Nullable List<String> onTabComplete(
      @NotNull CommandSender sender,
      @NotNull Command command,
      @NotNull String label,
      @NotNull String[] args) {
    if (args.length == 1) {
      return List.of("create", DELETE, "list", MODIFY, "save", DISPLAY);
    } else if (args.length == 2
        && (args[0].equalsIgnoreCase(DELETE)
            || args[0].equalsIgnoreCase(MODIFY)
            || args[0].equalsIgnoreCase(DISPLAY))) {
      ClientBuildManager clientBuildManager = Manager.getInstance(ClientBuildManager.class);
      List<String> buildNames = clientBuildManager.getBuildNames();
      if (nameCompletions.isEmpty() || !new HashSet<>(nameCompletions).containsAll(buildNames)) {
        nameCompletions = buildNames;
      }
      return nameCompletions;
    }
    return null;
  }
}
