package me.unurled.sacredrealms.sr.components.treasure;

import static me.unurled.sacredrealms.sr.utils.Component.textComp;

import com.google.gson.Gson;
import com.google.gson.GsonBuilder;
import io.papermc.paper.event.packet.PlayerChunkLoadEvent;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import javax.annotation.Nullable;
import me.unurled.sacredrealms.sr.components.player.PlayerManager;
import me.unurled.sacredrealms.sr.components.player.SRPlayer;
import me.unurled.sacredrealms.sr.data.DataHandler;
import me.unurled.sacredrealms.sr.data.DataManager;
import me.unurled.sacredrealms.sr.managers.Manager;
import org.bukkit.Chunk;
import org.bukkit.Location;
import org.bukkit.Material;
import org.bukkit.Sound;
import org.bukkit.block.Block;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.inventory.InventoryCloseEvent;
import org.bukkit.event.player.PlayerInteractEvent;
import org.bukkit.inventory.ItemStack;
import org.jetbrains.annotations.NotNull;

public class TreasureManager extends Manager {
  private final Map<Integer, Treasure> treasures = new HashMap<>();
  private final Map<Integer, Location> locations = new HashMap<>();

  private int nextId = 0;

  /** Save the data */
  @Override
  public void saveData() {
    // save all treasures
    for (Treasure treasure : treasures.values()) {
      // serialize treasure
      DataHandler dh = Manager.getInstance(DataManager.class).getDataHandler();
      Gson gson =
          new GsonBuilder().registerTypeAdapter(Treasure.class, new TreasureSerializer()).create();
      dh.set("treasures." + treasure.getId(), gson.toJson(treasure));
    }
  }

  /** Load the data */
  @Override
  public void loadData() {
    // load all treasures
    DataHandler dh = Manager.getInstance(DataManager.class).getDataHandler();
    Gson gson =
        new GsonBuilder().registerTypeAdapter(Treasure.class, new TreasureDeserializer()).create();
    for (String key : dh.getKeysAll("treasures")) {
      // deserialize treasure
      addTreasure(gson.fromJson(dh.get("treasures." + key), Treasure.class));
    }
  }

  public void addTreasure(Treasure treasure) {
    treasures.put(treasure.getId(), treasure);
    nextId = Math.max(nextId, treasure.getId() + 1);
    locations.put(treasure.getId(), treasure.getLocation());
  }

  @Nullable
  private Treasure getTreasureById(Integer id) {
    return treasures.get(id);
  }

  @Nullable
  public Treasure findByLocation(@NotNull Location loc) {
    for (Entry<Integer, Location> location : locations.entrySet()) {
      if (location.getValue().equals(loc)) {
        return getTreasureById(location.getKey());
      }
    }
    return null;
  }

  @Nullable
  private Treasure findByNearbyPlayer(@NotNull Location loc) {
    for (Entry<Integer, Location> location : locations.entrySet()) {
      if (location.getValue().getWorld() != loc.getWorld()) {
        continue;
      }
      if (location.getValue().distance(loc) < 5) {
        return getTreasureById(location.getKey());
      }
    }
    return null;
  }

  @NotNull
  public List<Treasure> getTreasures() {
    return treasures.values().stream().toList();
  }

  public void removeTreasure(@NotNull Treasure treasure) {
    Chunk chunk = treasure.getLocation().getChunk();
    Arrays.stream(chunk.getEntities())
        .map(
            entity -> {
              if (entity instanceof Player p) {
                return p;
              }
              return null;
            })
        .forEach(
            player -> {
              if (player != null) {
                player.sendBlockChange(treasure.getLocation(), Material.AIR.createBlockData());
              }
            });
    treasures.remove(treasure.getId());
  }

  public int getNextId() {
    return nextId++;
  }

  @EventHandler
  public void onChestOpen(@NotNull PlayerInteractEvent e) {
    if (e.hasBlock()) {
      Block clickedBlock = e.getClickedBlock();
      if (clickedBlock == null
          || !clickedBlock.getBlockData().getMaterial().equals(Material.CHEST)) {
        return;
      }

      onTreasureOpen(e, clickedBlock);
    }
  }

  private void onTreasureOpen(PlayerInteractEvent e, Block clickedBlock) {
    // it is a chest, lets loop on all treasure and find if the clicked block is a treasure chest

    for (Entry<Integer, Location> location : locations.entrySet()) {
      if (location.getValue().equals(clickedBlock.getLocation())) {
        // it is a treasure chest, lets open it
        Player p = e.getPlayer();
        Treasure treasure = getTreasureById(location.getKey());
        if (treasure == null) {
          p.sendMessage(textComp("<red>An Error occurred contact an Admin."));
          return;
        }
        if (!p.hasPermission(treasure.getPermission())) {
          p.sendMessage("You don't have permission to open this chest");
          p.playSound(p.getLocation(), Sound.BLOCK_ANVIL_BREAK, 1f, 0.5f);
          e.setCancelled(true);
          return;
        }
        e.setCancelled(true);
        p.openInventory(treasure.getInventory());
        SRPlayer srPlayer = Manager.getInstance(PlayerManager.class).getPlayer(p.getUniqueId());
        if (srPlayer != null) {
          srPlayer.addTreasureOpened(treasure.getId());
        }
        break;
      }
    }
  }

  @EventHandler
  public void onCloseTreasure(@NotNull InventoryCloseEvent e) {
    if (e.getView().title().contains(textComp("Treasure"))) {
      Treasure treasure = findByNearbyPlayer(e.getPlayer().getLocation());
      if (treasure != null) {
        // drop all items to the ground
        for (int i = 0; i < treasure.getInventory().getSize(); i++) {
          ItemStack item = treasure.getInventory().getItem(i);
          if (item != null) {
            e.getPlayer().getWorld().dropItemNaturally(e.getPlayer().getLocation(), item);
          }
        }
        Player p = (Player) e.getPlayer();
        p.sendBlockChange(treasure.getLocation(), Material.AIR.createBlockData());
      }
    }
  }

  @EventHandler
  public void onChunkLoad(@NotNull PlayerChunkLoadEvent e) {
    Player p = e.getPlayer();
    Treasure treasure = findByNearbyPlayer(p.getLocation());
    if (treasure != null) {
      SRPlayer srPlayer = Manager.getInstance(PlayerManager.class).getPlayer(p.getUniqueId());
      if (srPlayer != null && srPlayer.hasOpenedTreasure(treasure.getId())) {
        return;
      }
      // display chest to clientside builds
      p.sendBlockChange(treasure.getLocation(), Material.CHEST.createBlockData());
    }
  }
}
